#ifndef __IMAGE3D__
#define __IMAGE3D__

#include <GlewOpenGL.h>
#include <QImage>
#include "TextureData.h"
#include "ColorTypes.h"
#include "Texture.h"

namespace apig {

    /** This class is an abstract representation for three-dimensional images. */
    class AbstractImage3D : public TextureData {
        public:
            AbstractImage3D(int w=0, int h=0, int d=0) : w(w), h(h), d(d) {}
            virtual ~AbstractImage3D() {}
    
            virtual bool loaded() const = 0;
            virtual void destroy() = 0;
    
            int width()  const { return w; }
            int height() const { return h; }
            int depth()  const { return d; }
            bool contains(int i, int j, int k)       const { return (i >= 0) && (i < w) && (j >= 0) && (j < h) && (k >= 0) && (k < d); }
            bool contains(float x, float y, float z) const { return (x >= 0) && (x < w) && (y >= 0) && (y < h) && (z >= 0) && (z < d); }
            
            virtual GLint defaultTexFormat() const = 0;
            virtual GLenum textureMode() const { return GL_TEXTURE_3D; }
            virtual void loadToGPU(GLint texFormat) const { loadTexture3D(texFormat, GL_TEXTURE_3D); }
            virtual void loadTexture3D(GLint texFormat, GLenum target = GL_TEXTURE_3D) const = 0;
    
            enum WrapMode { CLAMP_TO_EDGE, CLAMP_TO_BORDER, REPEAT, MIRRORED_REPEAT };
            
        protected:
            int w, h, d;
        };
    
    /////////////////////////////////////////////////
    
    /** This class represents a three-dimensional image.
     * It is parameterized by the data type representing colors (number of channels and type used per channel).
     * Useful types for color representations are defined in \c ColorTypes.h : \c Float1, \c Float2, \c Float3, \c Float4, \c UByte3, \c UByte4, \c UInt3, \c UInt4.
     */
    template<class Color>
    class Image3D : public AbstractImage3D {
        public:
            Image3D(int w=0, int h=0, int d=0, Color* data=NULL);
            Image3D(QString fileName);
            virtual ~Image3D() {}
    
            virtual bool loaded() const { return data != NULL; }
            virtual void destroy();
            Image3D<Color> clone() const;
            void save(QString dirName, QString subDirName) const;
            
            virtual GLint defaultTexFormat() const { return Color::TEX_FORMAT; }
            virtual void loadTexture3D(GLint texFormat, GLenum target = GL_TEXTURE_3D) const;
    
            // acquisition d'une texture 3D :
            static Image3D<Color> readTexture(Texture *tex);
            
            // definition de l'echantillonnage aux bords de l'image :
            void setBorderColor(Color border);
            void setWrapMode(WrapMode wrapMode);
            void setupBorder(WrapMode wrapMode, Color border);
            
            // fonctions d'acces aux pixels :
            //-------------------------------
            inline       Color& texel(int i, int j, int k)       { return data[i + (j + k*h)*w]; }    // precondition : this->contains(i,j,k)
            inline const Color& texel(int i, int j, int k) const { return data[i + (j + k*h)*w]; }    // precondition : this->contains(i,j,k)
            inline       Color& operator()(int i, int j, int k)       { return texel(i,j,k); }
            inline const Color& operator()(int i, int j, int k) const { return texel(i,j,k); }
            Color sample(int i, int j, int k) const;                // si !this->contains(i,j,k), le resultat depend de <borderColor> et de <wrapMode>
            Color interp(float x, float y, float z) const;          // interpolation trilineaire : image mappee sur [0,w[*[0,h[*[0,d[
            Color operator()(float x, float y, float z) const { return interp(x,y,z); }
            
        private:
            Color *data;
            Color borderColor;
            WrapMode wrapMode;
        };
    
    typedef Image3D<UByte4> Image3DUByte4;
    typedef Image3D<Float1> Image3DFloat;
    typedef Image3D<Float3> Image3DFloat3;
    typedef Image3D<Float4> Image3DFloat4;
    
    }

#include "Image3D_impl.h"

#endif

