#include "Texture.h"
#include "CubeMap.h"
#include "Message.h"
//#include <GL/glprocs.h>

namespace apig {

Texture::Texture(const TextureData *texData, GLenum interpMode, GLenum wrapMode)
        : id(0),
          texData(texData),
          texMode(texData->textureMode()),
          interpMode(interpMode),
          wrapMode(wrapMode),
          loaded(false) {
    }

Texture::Texture(GLenum texMode, GLenum interpMode, GLenum wrapMode)
        : id(0),
          texData(NULL),
          texMode(texMode),
          interpMode(interpMode),
          wrapMode(wrapMode),
          loaded(false) {
    generateID();
    }

/***********************************************************************************************/

bool Texture::hasContent() const {
    switch (texMode) {
        case GL_TEXTURE_1D : return getWidth() > 0;
        case GL_TEXTURE_2D : return getWidth() > 0 && getHeight() > 0;
        case GL_TEXTURE_3D : return getWidth() > 0 && getHeight() > 0 && getDepth() > 0;
        };
    // cas restant : GL_TEXTURE_CUBE_MAP :
    bool defined = true;
    bind();
    for (int i=0; i<6; i++) {
        int w, h;
        glGetTexLevelParameteriv(CubeMap<Float4>::targetNames[i], 0, GL_TEXTURE_WIDTH, &w);
        glGetTexLevelParameteriv(CubeMap<Float4>::targetNames[i], 0, GL_TEXTURE_HEIGHT, &h);
        defined = defined && w > 0 && h > 0;
        }
    return defined;
    }

/***********************************************************************************************/

void Texture::load() {
    if (texData == NULL) {
        Message::error("pas de donnees associees a la texture");
        return;
        }
    load(texData->defaultTexFormat());
    }

void Texture::load(GLint internalFormat) {
    if (texData == NULL) {
        Message::error("pas de donnees associees a la texture");
        return;
        }
    if (id <= 0) generateID();
    init();
    texData->loadToGPU(internalFormat);
    loaded = true;
    }

/***********************************************************************************************/

void Texture::destroy() {
    if (id == 0) return;
    glDeleteTextures(1, &id);
    id = 0;
    loaded = false;
    }

/***********************************************************************************************/

void Texture::updateInterpMode() const {
    glTexParameteri(texMode, GL_TEXTURE_MAG_FILTER, interpMode);
    glTexParameteri(texMode, GL_TEXTURE_MIN_FILTER, interpMode);
    }

void Texture::updateWrapMode() const {
    glTexParameteri(texMode, GL_TEXTURE_WRAP_S, wrapMode);
    glTexParameteri(texMode, GL_TEXTURE_WRAP_T, wrapMode);
    glTexParameteri(texMode, GL_TEXTURE_WRAP_R, wrapMode);
    }

void Texture::setInterpMode(GLenum interpMode) {
    this->interpMode = interpMode;
    bind();
    updateInterpMode();
    }

void Texture::setWrapMode(GLenum wrapMode) {
    this->wrapMode = wrapMode;
    bind();
    updateWrapMode();
    }

void Texture::setBorderColor(const float *color) const {
    bind();
    glTexParameterfv(texMode, GL_TEXTURE_BORDER_COLOR, color);
    }

void Texture::setBorderColor(float r, float g, float b, float a) const {
    const float color[4] = { r, g, b, a };
    setBorderColor(color);
    }

/***********************************************************************************************/

void Texture::init() const {
    bind();
    updateInterpMode();
    updateWrapMode();
    }

void Texture::generateID() {
    glGenTextures(1, &id);
    if (id <= 0)
        Message::error("probleme lors de la generation de l'identifiant de texture");
    }

void Texture::bind() const {
    glBindTexture(texMode, id);
    }

void Texture::activate() const {
    glEnable(texMode);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    bind();
    }

void Texture::deactivate() const {
    glDisable(texMode);
    }

void Texture::activate(int texNum) const {
    glActiveTextureARB(GL_TEXTURE0_ARB + texNum);
    glEnable(texMode);
    bind();
    }

void Texture::deactivate(int texNum) const {
    glActiveTextureARB(GL_TEXTURE0_ARB + texNum);
    glDisable(texMode);
    }

/***********************************************************************************************/

Texture Texture::createTex1D(int w, GLint internalFormat, GLenum interpMode, GLenum wrapMode) {
    Texture tex(GL_TEXTURE_1D, interpMode, wrapMode);
    tex.init();
    glTexImage1D(GL_TEXTURE_1D, 0, internalFormat, w, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);   // arguments 7 et 8 quelconques
    tex.loaded = true;
    return tex;
    }

Texture Texture::createTex2D(int w, int h, GLint internalFormat, GLenum interpMode, GLenum wrapMode) {
    Texture tex(GL_TEXTURE_2D, interpMode, wrapMode);
    tex.init();
    glTexImage2D(GL_TEXTURE_2D, 0, internalFormat, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);   // arguments 7 et 8 quelconques
    tex.loaded = true;
    return tex;
    }

Texture Texture::createTex3D(int w, int h, int d, GLint internalFormat, GLenum interpMode, GLenum wrapMode) {
    Texture tex(GL_TEXTURE_3D, interpMode, wrapMode);
    tex.init();
    glTexImage3D(GL_TEXTURE_3D, 0, internalFormat, w, h, d, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);   // arguments 7 et 8 quelconques
    tex.loaded = true;
    return tex;
    }

Texture Texture::createTexCube(int n, GLint internalFormat, GLenum interpMode, GLenum wrapMode) {
    Texture tex(GL_TEXTURE_CUBE_MAP, interpMode, wrapMode);
    tex.init();
    for (int i=0; i<6; i++)
        glTexImage2D(CubeMap<Float4>::targetNames[i], 0, internalFormat, n, n, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);   // arguments 7 et 8 quelconques
    tex.loaded = true;
    return tex;
    }

/***********************************************************************************************/

int Texture::getParam(GLenum paramName) const {
    int res;
    bind();
    glGetTexParameteriv(texMode, paramName, &res);
    return res;
    }

int Texture::getLevelParam(int target, GLenum paramName, int level) const {
    int res;
    bind();
    glGetTexLevelParameteriv(target, level, paramName, &res);
    return res;
    }

Dim2D Texture::renderSize(GLenum target, int level) const {
    int w = getLevelParam(target, GL_TEXTURE_WIDTH,  level),
        h = getLevelParam(target, GL_TEXTURE_HEIGHT, level);
    return Dim2D(w, h);
    }

}

