#ifndef __VEC2__
#define __VEC2__

#include <QString>
#include <QDomElement>
#include <QDomDocument>


namespace apig {
    class Vec3;
    
    /** Representation of a two-dimensional vector. */
    class Vec2 {
        public:
            Vec2(float x, float y);
            Vec2(float c = 0);      // vecteur (c,c)
            Vec2(const float *v);   // tableau de 2 valeurs
            Vec2(const QDomElement &element);
            
        // operateurs de conversion :
        //---------------------------
            // operateurs de conversion automatique :
            inline operator const float* () const { return v; }     // cast en const float*
            inline operator       float* ()       { return v; }     // cast en float*
            // autres conversions :
            QString toQString() const;
            // lecture / ecriture d'un element XML :
            void initFromDOMElement(const QDomElement &element);
            QDomElement domElement(const QString &name, QDomDocument &document) const;
    
        // operations geometriques :
        //--------------------------
            float norm() const;         // norme
            float norm2() const;        // norme au carre
            void  normalize();          // normalise le vecteur
            Vec2  normalized() const;   // retourne le vecteur normalise
    
            friend float dist(Vec2 v1, Vec2 v2);    // distance euclidienne
            
            friend float dot(const Vec2 &a, const Vec2 &b);     // produit scalaire
            friend Vec3  vec(const Vec2 &a, const Vec2 &b);     // produit vectoriel
            friend float vecz(const Vec2 &a, const Vec2 &b);    // composante z du produit vectoriel
            friend float operator|(const Vec2 &a, const Vec2 &b);   // produit scalaire
            friend Vec3  operator^(const Vec2 &a, const Vec2 &b);   // produit vectoriel
    
            float arg() const;      // retourne l'angle par rapport a l'axe des abscisses (dans ]-Pi,Pi])
            Vec2 polar() const;     // retourne les coordonnees polaires (r, theta)
            
        // operateurs arithmetiques :
        //---------------------------
            // operateurs arithmetiques unaires :
            friend Vec2 operator-(const Vec2 &a);
    
            // operateurs arithmetiques binaires :
            friend Vec2 operator+(const Vec2 &a, const Vec2 &b);
            friend Vec2 operator-(const Vec2 &a, const Vec2 &b);
            friend Vec2 operator*(const Vec2 &a, const Vec2 &b);
            friend Vec2 operator/(const Vec2 &a, const Vec2 &b);
                
            // operations avec un scalaire :
            friend Vec2 operator*(float s, const Vec2 &a);
            friend Vec2 operator*(const Vec2 &a, float s);
            friend Vec2 operator/(float s, const Vec2 &a);
            friend Vec2 operator/(const Vec2 &a, float s);
            
            // operateurs de modification :
            Vec2& operator+=(const Vec2 &a);
            Vec2& operator-=(const Vec2 &a);
            Vec2& operator*=(const Vec2 &a);
            Vec2& operator/=(const Vec2 &a);
            Vec2& operator*=(float s);
            Vec2& operator/=(float s);
        
        // operateurs de comparaison :
        //----------------------------
            friend bool operator==(const Vec2 &a, const Vec2 &b);   // comparaison a 1e-10 pres
            friend bool operator!=(const Vec2 &a, const Vec2 &b);   // comparaison a 1e-10 pres
            // les comparaisons suivantes sont le <et> booleen des comparaisons sur les composantes :
            friend bool operator>=(const Vec2 &a, const Vec2 &b);
            friend bool operator<=(const Vec2 &a, const Vec2 &b);
            friend bool operator>(const Vec2 &a, const Vec2 &b);
            friend bool operator<(const Vec2 &a, const Vec2 &b);
            friend bool operator>=(const Vec2 &a, float b);
            friend bool operator<=(const Vec2 &a, float b);
            friend bool operator>(const Vec2 &a, float b);
            friend bool operator<(const Vec2 &a, float b);
        
        // autres fonctions pratiques :
        //-----------------------------
            friend Vec2 abs(const Vec2 &a);
            friend Vec2 sign(const Vec2 &a);
            friend Vec2 floor(const Vec2 &a);
            friend Vec2 ceil(const Vec2 &a);
            friend Vec2 fract(const Vec2 &a);
            friend Vec2 min(const Vec2 &a, const Vec2 &b);
            friend Vec2 min(const Vec2 &a, float b);
            friend Vec2 max(const Vec2 &a, const Vec2 &b);
            friend Vec2 max(const Vec2 &a, float b);
            friend Vec2 clamp(const Vec2 &a, const Vec2 &min, const Vec2 &max);
            friend Vec2 clamp(const Vec2 &a, float min, float max);
            friend Vec2 mix(const Vec2 &a, const Vec2 &b, const Vec2 &alpha);     // (1-alpha) * a + alpha * b
            friend Vec2 mix(const Vec2 &a, const Vec2 &b, float alpha);           // (1-alpha) * a + alpha * b
            friend Vec2 step(const Vec2 &e, const Vec2 &a);
            friend Vec2 step(float e, const Vec2 &a);
            friend float min(const Vec2 &a);    // minimum des composantes
            friend float max(const Vec2 &a);    // maximum des composantes
        
        // calcul de vecteurs aleatoires (variables uniformes) :
        //------------------------------------------------------
            static Vec2 random();                       // valeurs dans [0,1[
            static Vec2 random(float min, float max);   // valeurs dans [min,max[
            static Vec2 random(Vec2 min, Vec2 max);     // bornes par composantes
    
        // operations OpenGL :
        //--------------------
            void glVertex() const;    // effectue le glVertex correspondant
    
        public:
            union {
                struct { float x, y; };
                float v[2];
                };
        };
    
    }

#endif

